from PyCompGeomAlgorithms.core import Point
from AlgoGrade.parsers import PointListGivenJSONParser


def test_point_list_parser():
    points = [Point(1, 1), Point(2, 2)]
    data = [{"coords": [1, 1]}, {"coords": [2, 2]}]

    assert PointListGivenJSONParser.parse(data) == (points,)


from copy import deepcopy
from functools import partial
from math import isclose
from pytest import fixture
from PyCompGeomAlgorithms.core import BinTree, BinTreeNode, ThreadedBinTree, ThreadedBinTreeNode
from AlgoGrade.core import Grader, Scoring


single_mistake_fine_scoring = Scoring(min_grade=0, max_grade=1, fine=0.5, repeat_fine=0)
repeated_mistake_fine_scoring = Scoring(min_grade=0, max_grade=1, fine=0.5, repeat_fine=1)


class ObjectGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [cls.grade_object]


class ApproximateEqualityObjectGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [partial(cls.grade_object, custom_eq=isclose)]


class IterableGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [cls.grade_iterable]


class BinTreeGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [cls.grade_bin_tree]


class ThreadedBinTreeGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [cls.grade_threaded_bin_tree]


def test_grade_object_regular_equality_correct_with_single_mistake_scoring():
    answers = [1]
    correct_answers = [1]

    total_score, _ = ObjectGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_object_regular_equality_correct_with_repeated_mistake_scoring():
    answers = [1]
    correct_answers = [1]

    total_score, _ = ObjectGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_object_regular_equality_incorrect_with_single_mistake_scoring():
    answers = [0]
    correct_answers = [1]

    total_score, _ = ObjectGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_object_regular_equality_incorrect_with_repeated_mistake_scoring():
    answers = [0]
    correct_answers = [1]

    total_score, _ = ObjectGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_object_custom_equality_with_single_mistake_scoring():
    answers = [1.33333333333]
    correct_answers = [1.333333333332]

    total_score, _ = ApproximateEqualityObjectGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_object_custom_equality_with_repeated_mistake_scoring():
    answers = [1.33333333333]
    correct_answers = [1.333333333332]

    total_score, _ = ApproximateEqualityObjectGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_iterable_correct_with_single_mistake_scoring():
    answers = [[1, 2, 3, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_iterable_correct_with_repeated_mistake_scoring():
    answers = [[1, 2, 3, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_iterablle_incorrect_single_with_single_mistake_scoring():
    answers = [[0, 2, 3, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_iterablle_incorrect_single_with_repeated_mistake_scoring():
    answers = [[0, 2, 3, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_iterable_incorrect_repeated_with_single_mistake_scoring():
    answers = [[0, 2, 0, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_iterable_incorrect_repeated_with_repeated_mistake_scoring():
    answers = [[0, 2, 0, 4]]
    correct_answers = [[1, 2, 3, 4]]

    total_score, _ = IterableGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_bin_tree_correct_with_single_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_bin_tree_correct_with_repeated_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_bin_tree_incorrect_contents_single_with_single_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(0), BinTreeNode(3)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_incorrect_contents_single_with_repeated_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(0), BinTreeNode(3)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_incorrect_contents_repeated_with_single_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(0), BinTreeNode(0)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_incorrect_contents_repeated_with_repeated_mistake_scoring():
    answers = [BinTree(BinTreeNode(1, BinTreeNode(0), BinTreeNode(0)))]
    correct_answers = [BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_bin_tree_extra_nodes_single_with_single_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree.root.left.left = BinTreeNode("extra")

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_extra_nodes_single_with_repeated_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree.root.left.left = BinTreeNode("extra")

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_extra_nodes_repeated_with_single_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree.root.left.left = BinTreeNode("extra")
    tree.root.right.left = BinTreeNode("extra")
    tree.root.right.left.left = BinTreeNode("extra")

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_extra_nodes_repeated_with_repeated_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree.root.left.left = BinTreeNode("extra")
    tree.root.right.left = BinTreeNode("extra")
    tree.root.right.left.left = BinTreeNode("extra")

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_bin_tree_missing_nodes_single_with_single_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2)))

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_missing_nodes_single_with_repeated_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1, BinTreeNode(2)))

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_missing_nodes_repeated_with_single_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1))

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_bin_tree_missing_nodes_repeated_with_repeated_mistake_scoring():
    correct_tree = BinTree(BinTreeNode(1, BinTreeNode(2), BinTreeNode(3)))
    tree = BinTree(BinTreeNode(1))

    answers = [tree]
    correct_answers = [correct_tree]

    total_score, _ = BinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


@fixture
def correct_circular_tbt():
    root = ThreadedBinTreeNode(3)
    root.left = ThreadedBinTreeNode(0)
    root.left.right = ThreadedBinTreeNode(2)
    root.left.right.left = ThreadedBinTreeNode(1)
    root.right = ThreadedBinTreeNode(4)
    root.right.right = ThreadedBinTreeNode(5)

    root.prev = root.left.right
    root.next = root.right

    root.left.prev = root.right.right
    root.left.next = root.left.right.left

    root.left.right.prev = root.left.right.left
    root.left.right.next = root

    root.left.right.left.prev = root.left
    root.left.right.left.next = root.left.right

    root.right.prev = root
    root.right.next = root.right.right

    root.right.right.prev = root.right
    root.right.right.next = root.left

    return ThreadedBinTree(root)


@fixture
def correct_non_circular_tbt(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.prev = None
    tbt.root.right.right.next = None

    return tbt


def test_grade_threaded_bin_tree_correct_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_threaded_bin_tree_correct_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 1)


def test_grade_threaded_bin_tree_incorrect_contents_single_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.data = 100

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_incorrect_contents_single_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.data = 100

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_incorrect_contents_repeated_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.data = 100
    tbt.root.right.data = 100

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_incorrect_contents_repeated_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.data = 100
    tbt.root.right.data = 100

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_threaded_bin_tree_extra_nodes_single_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.left = ThreadedBinTreeNode("extra1")

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_nodes_single_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.left = ThreadedBinTreeNode("extra1")

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_nodes_repeated_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.left = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right = ThreadedBinTreeNode("extra2")

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_nodes_repeated_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.left = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right = ThreadedBinTreeNode("extra2")

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_threaded_bin_tree_missing_nodes_single_with_single_mistake_scoring(correct_circular_tbt):
    # Note that this test also triggers missing threads by necessity, so there are actually 3 mistakes here (1 for node, 2 for threads)
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.right.left = None
    tbt.root.left.right.prev = None
    tbt.root.left.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_missing_nodes_single_with_repeated_mistake_scoring(correct_circular_tbt):
    # Note that this test also triggers missing threads by necessity, so there are actually 3 mistakes here (1 for node, 2 for threads)
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.right.left = None
    tbt.root.left.right.prev = None
    tbt.root.left.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.0)


def test_grade_threaded_bin_tree_missing_nodes_repeated_with_single_mistake_scoring(correct_circular_tbt):
    # Note that this test also triggers missing threads by necessity, so there are actually 6 mistakes here (2 for nodes, 4 for threads)
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.right.left = None
    tbt.root.left.right.prev = None
    tbt.root.left.next = None

    tbt.root.right.right = None
    tbt.root.left.prev = None
    tbt.root.right.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_missing_nodes_repeated_with_repeated_mistake_scoring(correct_circular_tbt):
    # Note that this test also triggers missing threads by necessity, so there are actually 6 mistakes here (2 for nodes, 4 for threads)
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.left.right.left = None
    tbt.root.left.right.prev = None
    tbt.root.left.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.0)


def test_grade_threaded_bin_tree_extra_threads_in_incorrect_part_of_tree_single_with_single_mistake_scoring(correct_non_circular_tbt):
    # Note that this test also triggers extra nodes by necessity, so there are actually 2 mistakes here (1 for node, 1 for thread)
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.right.right = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left.right.right

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_threads_in_incorrect_part_of_tree_single_with_repeated_mistake_scoring(correct_non_circular_tbt):
    # Note that this test also triggers extra nodes by necessity, so there are actually 2 mistakes here (1 for node, 1 for thread)
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.right.right = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left.right.right

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.0)


def test_grade_threaded_bin_tree_extra_threads_in_incorrect_part_of_tree_repeated_with_single_mistake_scoring(correct_non_circular_tbt):
    # Note that this test also triggers extra nodes by necessity, so there are actually 4 mistakes here (2 for nodes, 2 for threads)
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.right.right = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left.right.right

    tbt.root.left.right.right.right = ThreadedBinTreeNode("extra2")
    tbt.root.left.right.right.next = tbt.root.left.right.right.right
    tbt.root.left.right.right.right.prev = tbt.root.left.right.right

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_threads_in_incorrect_part_of_tree_repeated_with_repeated_mistake_scoring(correct_non_circular_tbt):
    # Note that this test also triggers extra nodes by necessity, so there are actually 4 mistakes here (2 for nodes, 2 for threads)
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.right.right = ThreadedBinTreeNode("extra1")
    tbt.root.left.right.right.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left.right.right

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.0)


def test_grade_threaded_bin_tree_missing_threads_in_correct_subtree_single_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.prev = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_missing_threads_in_correct_subtree_single_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.prev = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_missing_threads_in_correct_subtree_repeated_with_single_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.prev = None
    tbt.root.left.right.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_missing_threads_in_correct_subtree_repeated_with_repeated_mistake_scoring(correct_circular_tbt):
    tbt = deepcopy(correct_circular_tbt)
    tbt.root.prev = None
    tbt.root.left.right.next = None

    answers = [tbt]
    correct_answers = [correct_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_threaded_bin_tree_extra_thread_in_correct_subtree_with_single_mistake_scoring(correct_non_circular_tbt):
    # There can be only one extra thread in a correct subtree, and it's the one that connects the extreme nodes in a non-circular TBT (which shouldn't be connected).
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_extra_thread_in_correct_subtree_with_repeated_mistake_scoring(correct_non_circular_tbt):
    # There can be only one extra thread in a correct subtree, and it's the one that connects the extreme nodes in a non-circular TBT (which shouldn't be connected).
    tbt = deepcopy(correct_non_circular_tbt)
    tbt.root.left.prev = tbt.root.right.right
    tbt.root.right.right.next = tbt.root.left

    answers = [tbt]
    correct_answers = [correct_non_circular_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


@fixture
def correct_shallow_tbt():
    root = ThreadedBinTreeNode(1)
    root.left = ThreadedBinTreeNode(0)
    root.right = ThreadedBinTreeNode(2)

    root.prev = root.left
    root.next = root.right

    root.left.prev = root.right
    root.left.next = root

    root.right.prev = root
    root.right.next = root.left

    return ThreadedBinTree(root)


@fixture
def incorrect_shallow_tbt():
    root = ThreadedBinTreeNode(1)
    root.left = ThreadedBinTreeNode(0)
    root.right = ThreadedBinTreeNode(2)

    root.prev = root.right
    root.right.next = root

    root.left.next = None
    root.left.prev = None

    return ThreadedBinTree(root)


def test_grade_threaded_bin_tree_threads_in_correct_subtree_in_wrong_direction_single_with_single_mistake_scoring(correct_shallow_tbt, incorrect_shallow_tbt):
    # Note that this test also triggers missing threads in correct subtree by necessity, so there are actually 3 mistakes (1 thread in wrong direction and 2 missing threads)
    answers = [incorrect_shallow_tbt]
    correct_answers = [correct_shallow_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_threads_in_correct_subtree_in_wrong_direction_single_with_repeated_mistake_scoring(correct_shallow_tbt, incorrect_shallow_tbt):
    # Note that this test also triggers missing threads in correct subtree by necessity, so there are actually 3 mistakes (1 thread in wrong direction and 2 missing threads)
    answers = [incorrect_shallow_tbt]
    correct_answers = [correct_shallow_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


def test_grade_threaded_bin_tree_threads_in_correct_subtree_in_wrong_direction_repeated_with_single_mistake_scoring(correct_shallow_tbt):
    tbt = deepcopy(correct_shallow_tbt)
    tbt.root.prev = tbt.root.right
    tbt.root.next = tbt.root.left

    tbt.root.left.prev = tbt.root
    tbt.root.left.next = tbt.root.right

    tbt.root.right.prev = tbt.root.left
    tbt.root.right.next = tbt.root
    
    answers = [tbt]
    correct_answers = [correct_shallow_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [single_mistake_fine_scoring])
    assert isclose(total_score, 0.5)


def test_grade_threaded_bin_tree_threads_in_correct_subtree_in_wrong_direction_repeated_with_repeated_mistake_scoring(correct_shallow_tbt):
    tbt = deepcopy(correct_shallow_tbt)
    tbt.root.prev = tbt.root.right
    tbt.root.next = tbt.root.left

    tbt.root.left.prev = tbt.root
    tbt.root.left.next = tbt.root.right

    tbt.root.right.prev = tbt.root.left
    tbt.root.right.next = tbt.root
    
    answers = [tbt]
    correct_answers = [correct_shallow_tbt]

    total_score, _ = ThreadedBinTreeGrader.grade_pycga(answers, correct_answers, [repeated_mistake_fine_scoring])
    assert isclose(total_score, 0)


from collections import Counter
from functools import partial
from operator import eq
from typing import Iterable, Callable, Type, Any
from pydantic import BaseModel
from .adapters import pycga_to_pydantic, pydantic_to_pycga, SerializablePydanticModelWithPydanticFields


class Answers(SerializablePydanticModelWithPydanticFields):
    @classmethod
    def from_iterable(cls, iterable):
        raise NotImplementedError
    
    def to_pydantic_list(self):
        raise NotImplementedError
    
    def to_pycga_list(self):
        return pydantic_to_pycga(self.to_pydantic_list())


class Mistake:
    def __init__(self, scorings, description=""):
        self.scorings = scorings
        self.description = description
    
    @property
    def is_repeated(self):
        return self.scorings.repeat_fine > 0
    
    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.scorings == other.scorings
        )
    
    def __hash__(self):
        return hash((self.__class__, self.scorings))


class Scoring(BaseModel, frozen=True):
    min_grade: float = -1000.0
    max_grade: float = 0.0
    fine: float = 0.0
    repeat_fine: float = 0.0


class Grader:
    @classmethod
    def grade_methods(cls):
        """List of grading methods for each step of the task. Redefine in the derived class"""
        raise NotImplementedError

    @classmethod
    def grade_pycga(cls, answers, correct_answers, scorings):
        """
            Compare answers to correct_answers (both in PyCGA format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PyCGA format.
        """
        mistake_lists = [
            grading_method(answer, correct_answer, scorings)
            for grading_method, answer, correct_answer, scorings
            in zip(cls.grade_methods(), answers, correct_answers, scorings)
        ]
        mistake_counters = [Counter(mistakes) for mistakes in mistake_lists]
        mistake_fines_dicts = [
            {
                mistake: mistake.scorings.repeat_fine if mistake.is_repeated and count > 1 else mistake.scorings.fine
                for mistake, count in mistake_counter.items()
            }
            for mistake_counter in mistake_counters
        ]

        answers_grades = [
            (answer, max(scorings.min_grade, scorings.max_grade-sum(mistake_fine_dict.values())))
            for answer, scorings, mistake_fine_dict
            in zip(answers, scorings, mistake_fines_dicts)
        ]
        total_grade = sum(grade for answer, grade in answers_grades)
        
        return total_grade, answers_grades
    
    @classmethod
    def grade_pydantic(cls, answers, correct_answers, scorings):
        """
            Compare answers to correct_answers (both in PydanticAdapter format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PydanticAdapter format.
        """
        total_grade, answers_grades = cls.grade_pycga(pydantic_to_pycga(answers), pydantic_to_pycga(correct_answers), scorings)
        return total_grade, [(pycga_to_pydantic(answer), grade) for answer, grade in answers_grades]
    
    @classmethod
    def grade_answers_wrapper(cls, answers: Answers, correct_answers: Answers, scorings):
        """
            Compare answers to correct_answers (both in Answers format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PydanticAdapter format.
        """
        return cls.grade_pydantic(answers.to_pydantic_list(), correct_answers.to_pydantic_list(), scorings)
    
    @classmethod
    def grade_object(cls, answer, correct_answer, scorings, item_mistake_text="", custom_eq=eq):
        """
            The most basic method for grading an object compared to the correct one, using customizable equality.
            
            By default, grades `answer` compared to `correct_answer` based on whether `answer == correct_answer` (`answer.__eq__(correct_answer)`).

            Returns an empty list if the equality holds, and a list containing a mistake otherwise.
        """
        return [] if custom_eq(answer, correct_answer) else [Mistake(scorings, description=item_mistake_text)]

    @classmethod
    def grade_iterable(cls, answer, correct_answer, scorings, grade_item_method=None, item_mistake_text=""):
        if grade_item_method is None:
            grade_item_method = cls.grade_object
        if callable(grade_item_method):
            grade_item_method = [grade_item_method] * len(correct_answer)
        elif not isinstance(grade_item_method, Iterable):
            raise TypeError(f"grade_item_method should be either a grading method, or an iterable of grading methods for each item, or None (defaults to default_grading)")

        len_diff = len(correct_answer) - len(answer)
        if len_diff == 0:
            return flatten([
                g(a, c, scorings, item_mistake_text=item_mistake_text)
                for g, a, c in zip(grade_item_method, answer, correct_answer)
            ])

        return [Mistake(scorings, description=f"Too {'few' if len_diff < 0 else 'many'} items")] * abs(len_diff)

    @classmethod
    def grade_bin_tree(cls, answer, correct_answer, scorings, grade_item_method=None, item_mistake_text=""):
        if grade_item_method is None:
            grade_item_method = partial(cls.grade_object, custom_eq=lambda a, c: c.weak_equal(a))
        
        return cls._find_mistakes_in_bin_tree(answer.root, correct_answer.root, scorings, grade_item_method, item_mistake_text)

    @classmethod
    def _find_mistakes_in_bin_tree(
        cls, node, correct_node, scorings,
        grade_item_method=None, item_mistake_text="", mistakes=None, mistakes_extra=None, mistakes_missing=None
    ):
        if mistakes is None:
            mistakes = []
        if mistakes_extra is None:
            mistakes_extra = []
        if mistakes_missing is None:
            mistakes_missing = []
        
        mistakes.extend(grade_item_method(node, correct_node, scorings))

        cls._find_mistakes_in_subtree(node.left, correct_node.left, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing)
        cls._find_mistakes_in_subtree(node.right, correct_node.right, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing)

        return mistakes + mistakes_extra + mistakes_missing

    @classmethod
    def _find_mistakes_in_subtree(cls, node, correct_node, scorings, grade_item_method, item_mistake_text, mistakes, extra, missing):
        if node and correct_node:
            cls._find_mistakes_in_bin_tree(node, correct_node, scorings, grade_item_method, item_mistake_text, mistakes, extra, missing)
        if node and not correct_node:
            extra.extend(Mistake(scorings, "Extra item") for _ in node.traverse_inorder())
        if not node and correct_node:
            missing.extend(Mistake(scorings, "Missing item") for _ in correct_node.traverse_inorder())

    @classmethod
    def grade_threaded_bin_tree(cls, answer, correct_answer, scorings, grade_item_method=None, item_mistake_text=""):
        if grade_item_method is None:
            grade_item_method = partial(cls.grade_object, custom_eq=lambda a, c: c.weak_equal(a))

        nodes = answer.traverse_inorder()
        correct_nodes = correct_answer.traverse_inorder()

        cls._set_nodes_indices(nodes)
        cls._set_nodes_indices(correct_nodes)
        
        matching_nodes, correct_matching_nodes = [], []
        seen_threads = set()
        mistakes = cls._find_mistakes_in_threaded_bin_tree(
            node=answer.root, correct_node=correct_answer.root, scorings=scorings, grade_item_method=grade_item_method, item_mistake_text=item_mistake_text,
            matching_nodes=matching_nodes, correct_matching_nodes=correct_matching_nodes, seen_threads=seen_threads
        )

        for node, correct_node in zip(matching_nodes, correct_matching_nodes):
            mistakes.extend(cls._find_thread_mistakes_in_correct_subtree(node, node.prev, correct_node.prev, seen_threads, scorings))
            mistakes.extend(cls._find_thread_mistakes_in_correct_subtree(node, node.next, correct_node.next, seen_threads, scorings))
        
        return mistakes

    @classmethod
    def _set_nodes_indices(cls, nodes):
        for i, node in enumerate(nodes):
                node.matching_node_index = None
                node.inorder_index = i
        for node in nodes:
            if node.prev is not None:
                node.prev_index = node.prev.inorder_index
            else:
                node.prev_index = None
            if node.next is not None:
                node.next_index = node.next.inorder_index
            else:
                node.next_index = None

    @classmethod
    def _find_mistakes_in_threaded_bin_tree(
        cls, node, correct_node, scorings,
        grade_item_method=None, item_mistake_text="", mistakes=None, mistakes_extra=None, mistakes_missing=None,
        matching_nodes_counter=None, matching_nodes=None, correct_matching_nodes=None, seen_threads=None
    ):
        if mistakes is None:
            mistakes = []
        if mistakes_extra is None:
            mistakes_extra = []
        if mistakes_missing is None:
            mistakes_missing = []
        if matching_nodes_counter is None:
            matching_nodes_counter = [0]
        if matching_nodes is None:
            matching_nodes = []
        if correct_matching_nodes is None:
            correct_matching_nodes = []
        if seen_threads is None:
            seen_threads = set()

        mistakes.extend(grade_item_method(node, correct_node, scorings))
        
        matching_nodes_counter[0] += 1
        node.matching_node_index = matching_nodes_counter[0]
        correct_node.matching_node_index = matching_nodes_counter[0]
        
        matching_nodes.append(node)
        correct_matching_nodes.append(correct_node)

        cls._find_mistakes_in_threaded_subtree(node.left, correct_node.left, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing, matching_nodes_counter, matching_nodes, correct_matching_nodes, seen_threads)
        cls._find_mistakes_in_threaded_subtree(node.right, correct_node.right, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing, matching_nodes_counter, matching_nodes, correct_matching_nodes, seen_threads)
        
        return mistakes + mistakes_extra + mistakes_missing

    @classmethod
    def _find_mistakes_in_threaded_subtree(cls, subtree_root, correct_subtree_root, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing, matching_nodes_counter, matching_nodes, correct_matching_nodes, seen_threads):
        if subtree_root and correct_subtree_root:
            cls._find_mistakes_in_threaded_bin_tree(subtree_root, correct_subtree_root, scorings, grade_item_method, item_mistake_text, mistakes, mistakes_extra, mistakes_missing, matching_nodes_counter, matching_nodes, correct_matching_nodes, seen_threads)
        if subtree_root and not correct_subtree_root:
            extra_nodes = subtree_root.traverse_inorder()
            mistakes_extra.extend(Mistake(scorings, "Extra node") for _ in extra_nodes)            
            
            # Here threads can be set by student in an unpredictably wrong way, so we need to traverse them all to find their number
            connected_threads = cls._connected_threads(extra_nodes)
            mistakes_extra.extend(Mistake(scorings, "Extra thread") for _ in connected_threads)
            seen_threads |= connected_threads
        if not subtree_root and correct_subtree_root:
            missing_nodes = correct_subtree_root.traverse_inorder()
            mistakes_missing.extend(Mistake(scorings, "Missing node") for _ in missing_nodes)

            # Here the number of threads can be found without traversing.
            # If len(missing_nodes) is N > 0, then there is necessarily 1 thread between the parent of correct_subtree_root and the left/right extreme node of its (parent's) right/left subtree, respectively.
            # There are also N - 1 threads between N nodes in the subtree.
            # For circular TBT, there is 1 additional thread from that subtree's extreme node to either correct_node's parent or, if the extreme node is also globally extreme, to the other globally extreme node.
            # For non-circular TBT, there is no thread between globally leftmost node and globally rightmost node.
            n_missing_threads = len(missing_nodes) + (0 if missing_nodes[0].prev is None else 1)
            mistakes_missing.extend(Mistake(scorings, "Missing thread") for _ in range(n_missing_threads))

    @classmethod
    def _connected_threads(cls, nodes):
        seen_threads = set()
        for node in nodes:
            if node.prev_index is not None:
                seen_threads.add((node.prev_index, node.inorder_index))
            if node.next_index is not None:
                seen_threads.add((node.inorder_index, node.next_index))
        
        return seen_threads

    @classmethod
    def _find_thread_mistakes_in_correct_subtree(cls, node, neighbor, correct_neighbor, seen_threads, scorings):
        mistakes = []
        if neighbor is None and correct_neighbor is not None:
            mistakes.append(Mistake(scorings, "Missing thread"))
        elif neighbor is not None and (neighbor.inorder_index, node.inorder_index) not in seen_threads and (node.inorder_index, neighbor.inorder_index) not in seen_threads:
            if correct_neighbor is None:
                mistakes.append(Mistake(scorings, "Extra thread"))
                seen_threads.add((node.inorder_index, neighbor.inorder_index) if node.next is neighbor else (neighbor.inorder_index, node.inorder_index))
            else:
                threads_exist_and_dont_match = (
                    neighbor.matching_node_index is not None
                    and correct_neighbor.matching_node_index is not None
                    and neighbor.matching_node_index != correct_neighbor.matching_node_index
                )

                if threads_exist_and_dont_match:
                    mistakes.append(Mistake(scorings, "Thread doesn't match the correct one"))
                    seen_threads.add((node.inorder_index, neighbor.inorder_index) if node.next is neighbor else (neighbor.inorder_index, node.inorder_index))
                            
        return mistakes


class GivenJSONParser:
    """Parses a JSON-represented (list) givens as a tuple of arguments to be fed into the respective PyCGA task solving method."""
    @classmethod
    def parse(cls, data) -> list[Any]:
        """Parses a JSON-represented (list) givens as a tuple of arguments to be fed into the respective PyCGA task solving method."""
        raise NotImplementedError


class Task:
    algorithm: Callable = None
    grader_class: Type[Grader] = None
    answers_class: Type[Answers] = None
    given_parser_class: Type[GivenJSONParser] = None
    
    @classmethod
    def solve_as_pycga_list(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns a list of answers in PyCGA format."""
        return list(cls.algorithm(*givens))
    
    @classmethod
    def solve_as_pydantic_list(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns a list of answers in PydanticAdapter format."""
        pycga_answers = list(cls.solve_as_pycga_list(givens))
        return pycga_to_pydantic(pycga_answers)
    
    @classmethod
    def solve_as_answers_wrapper(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns answers in form of the respective Answers wrapper."""
        pydantic_answers = cls.solve_as_pydantic_list(givens)
        return cls.answers_class.from_iterable(pydantic_answers)


def flatten(iterable):
    return list(_flatten(iterable))


def _flatten(iterable):
    for item in iterable:
        if isinstance(item, Iterable) and not isinstance(item, str):
            for inner_item in flatten(item):
                yield inner_item
        else:
            yield item



from math import isclose
from copy import deepcopy
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.dynamic_hull import DynamicHullNode, DynamicHullTree, SubhullThreadedBinTree, PathDirection
from AlgoGrade.dynamic_hull import DynamicHullTask, DynamicHullGrader, DynamicHullAnswers
from AlgoGrade.adapters import pycga_to_pydantic
from AlgoGrade.core import Scoring


points = p2, p1, p3 = [Point(3, 3), Point(1, 1), Point(5, 0)]
point_to_insert = Point(4, 3)
givens = points, point_to_insert
task_class = DynamicHullTask
scorings = [
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.5, fine=0.25, repeat_fine=0.5),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.5, fine=0.25, repeat_fine=0.5),
    Scoring(max_grade=0.25, fine=0.25, repeat_fine=0.5),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.75, fine=0.25, repeat_fine=0.75)
]
correct_pycga_answers = task_class.solve_as_pycga_list(givens)
correct_pydantic_answers = task_class.solve_as_pydantic_list(givens)
correct_answers_wrapper = task_class.solve_as_answers_wrapper(givens)


def test_dynamic_hull_grader_all_correct():
    root = DynamicHullNode(p2, [p1, p2, p3], 1)
    root.left = DynamicHullNode(p1, [p1, p2])
    root.left.left = DynamicHullNode.leaf(p1)
    root.left.right = DynamicHullNode.leaf(p2)
    root.right = DynamicHullNode.leaf(p3)
    tree = DynamicHullTree(root)
    
    tree.root.optimized_subhull = tree.root.subhull
    tree.root.left.optimized_subhull = SubhullThreadedBinTree.empty()
    tree.root.left.left.optimized_subhull = SubhullThreadedBinTree.empty()
    tree.root.left.right.optimized_subhull = SubhullThreadedBinTree.empty()
    tree.root.right.optimized_subhull = SubhullThreadedBinTree.empty()
    
    leaves = [root.left.left, root.left.right, root.right]
    path = [PathDirection.right]
    hull = [p1, p2, point_to_insert, p3]

    modified_tree = deepcopy(tree)
    modified_tree.root.subhull = SubhullThreadedBinTree.from_iterable(hull)
    modified_tree.root.optimized_subhull = modified_tree.root.subhull
    modified_tree.root.right = DynamicHullNode(point_to_insert, [point_to_insert, p3])
    modified_tree.root.right.optimized_subhull = SubhullThreadedBinTree.empty()
    modified_tree.root.right.left = DynamicHullNode.leaf(point_to_insert)
    modified_tree.root.right.left.optimized_subhull = SubhullThreadedBinTree.empty()
    modified_tree.root.right.right = DynamicHullNode.leaf(p3)
    modified_tree.root.right.right.optimized_subhull = SubhullThreadedBinTree.empty()

    pycga_answers = [
        leaves,
        tree,
        tree,
        tree,
        tree,
        tree,
        path,
        (modified_tree, hull)
    ]
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 3)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 3)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 3)


def test_dynamic_hull_grader_incorrect_leaves():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[0] = deepcopy(pycga_answers[0])
    leaves = pycga_answers[0]
    leaves[0].data = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_left_supporting_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.left_supporting = Point(100, 100) # also triggers "omitted points" grading

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.5)



def test_dynamic_hull_grader_incorrect_right_supporting_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.right_supporting = Point(100, 100) # also triggers "omitted points" grading

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.5)


def test_dynamic_hull_grader_incorrect_left_and_right_supporting():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.left_supporting = Point(100, 100)  # also triggers "omitted points" grading
    pycga_answers[1].root.right_supporting = Point(100, 100) # also triggers "omitted points" grading

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


def test_dynamic_hull_grader_incorrect_left_supporting_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.left_supporting = Point(100, 100) # also triggers "omitted points" grading
    pycga_answers[1].root.left.left_supporting = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


def test_dynamic_hull_grader_incorrect_right_supporting_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.right_supporting = Point(100, 100) # also triggers "omitted points" grading
    pycga_answers[1].root.left.right_supporting = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


def test_dynamic_hull_grader_incorrect_omitted_points2():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[2].root.subhull = SubhullThreadedBinTree.from_iterable([Point(0, 0)] *3)
    
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75) # also triggers supporting points and subhull grading

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_dynamic_hull_grader_incorrect_subhull_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[3].root.left_supporting = pycga_answers[2].root.left_supporting
    pycga_answers[3].root.right_supporting = pycga_answers[2].root.right_supporting
    pycga_answers[3].root.subhull = deepcopy(pycga_answers[3].root.subhull) # to not interfere with optimized_subhull, which, in root, is a reference to subhull
    pycga_answers[3].root.subhull.root.point = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_subhull_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[3].root.left_supporting = pycga_answers[2].root.left_supporting
    pycga_answers[3].root.right_supporting = pycga_answers[2].root.right_supporting
    pycga_answers[3].root.subhull = deepcopy(pycga_answers[3].root.subhull) # to not interfere with optimized_subhull, which, in root, is a reference to subhull
    pycga_answers[3].root.subhull.root.point = Point(100, 100)

    pycga_answers[3].root.left.left_supporting = pycga_answers[2].root.left.left_supporting
    pycga_answers[3].root.left.right_supporting = pycga_answers[2].root.left.right_supporting
    pycga_answers[3].root.left.subhull.root.point = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.5)


def test_dynamic_hull_grader_incorrect_left_supporting_index_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[4].root.left_supporting = pycga_answers[2].root.left_supporting # getter returns this left_supporting, not subhull[left_supporting_index]
    pycga_answers[4].root.right_supporting = pycga_answers[2].root.right_supporting
    pycga_answers[4].root.left_supporting_index = 100

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_left_supporting_index_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[4].root.left_supporting = pycga_answers[2].root.left_supporting
    pycga_answers[4].root.right_supporting = pycga_answers[2].root.right_supporting
    pycga_answers[4].root.left_supporting_index = 100

    pycga_answers[4].root.left.left_supporting = pycga_answers[2].root.left.left_supporting
    pycga_answers[4].root.left.right_supporting = pycga_answers[2].root.left.right_supporting
    pycga_answers[4].root.left.left_supporting_index = 100

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.5)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.5)


def test_dynamic_hull_grader_incorrect_optimization():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[5].root.subhull = deepcopy(pycga_answers[5].root.subhull) # to prevent being modified when optimized_subhull is modified, which, in root, is a reference to subhull
    pycga_answers[5].root.optimized_subhull.root.point = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_search_path():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[6][0] = PathDirection.left

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_final_tree_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[7][0].root.point = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_final_hull_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[7][1][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.75)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.75)


def test_dynamic_hull_grader_incorrect_final_tree_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[7][0].root.point = Point(100, 100)
    pycga_answers[7][0].root.left.point = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


def test_dynamic_hull_grader_incorrect_final_hull_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[7][1][0] = Point(100, 100)
    pycga_answers[7][1][1] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


def test_dynamic_hull_grader_incorrect_final_tree_and_hull():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[7][0].root.point = Point(100, 100)
    pycga_answers[7][1][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = DynamicHullAnswers.from_iterable(pydantic_answers)

    total_grade, answers_grades = DynamicHullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2.25)

    total_grade, answers_grades = DynamicHullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2.25)


from __future__ import annotations
from functools import partial
from typing import ClassVar, Optional, Union
from PyCompGeomAlgorithms.core import BinTree
from PyCompGeomAlgorithms.quickhull import quickhull, QuickhullNode
from .adapters import pycga_to_pydantic, PointPydanticAdapter, BinTreeNodePydanticAdapter, BinTreePydanticAdapter
from .core import Task, Grader, Mistake, Answers
from .parsers import PointListGivenJSONParser


class QuickhullGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [
            cls.grade_iterable,
            partial(cls.grade_bin_tree, grade_item_method=lambda a, c, gp: cls.grade_object(a.h, c.h, gp)),
            partial(cls.grade_bin_tree, grade_item_method=lambda a, c, gp: cls.grade_iterable(a.points, c.points, gp)),
            cls.grade_finalization,
            partial(cls.grade_bin_tree, grade_item_method=lambda a, c, gp: cls.grade_iterable(a.subhull, c.subhull, gp))
        ]
    
    @classmethod
    def grade_finalization(cls, answer, correct_answer, scorings):
        return [Mistake(scorings) for node in answer.traverse_preorder() if not node.is_leaf and len(node.points) == 2]


class QuickhullNodePydanticAdapter(BinTreeNodePydanticAdapter):
    regular_class: ClassVar[type] = QuickhullNode
    data: list[PointPydanticAdapter]
    left: Optional[QuickhullNodePydanticAdapter] = None
    right: Optional[QuickhullNodePydanticAdapter] = None
    h: Optional[PointPydanticAdapter] = None
    subhull: Optional[list[PointPydanticAdapter]] = None

    @classmethod
    def from_regular_object(cls, obj: QuickhullNode, **kwargs):
        return super().from_regular_object(
            obj,
            h=pycga_to_pydantic(obj.h),
            subhull=pycga_to_pydantic(obj.subhull),
            **kwargs
        )


class QuickhullTreePydanticAdapter(BinTreePydanticAdapter):
    regular_class: ClassVar[type] = BinTree
    root: QuickhullNodePydanticAdapter


class QuickhullAnswers(Answers):
    leftmost_point: PointPydanticAdapter
    rightmost_point: PointPydanticAdapter
    subset1: list[PointPydanticAdapter]
    subset2: list[PointPydanticAdapter]
    tree: QuickhullTreePydanticAdapter

    @classmethod
    def from_iterable(cls, iterable):
        (leftmost_point, rightmost_point, subset1, subset2), tree, *rest = iterable
        return cls(
            leftmost_point=leftmost_point, rightmost_point=rightmost_point,
            subset1=subset1, subset2=subset2, tree=tree
        )
    
    def to_pydantic_list(self):
        return [
            (self.leftmost_point, self.rightmost_point, self.subset1, self.subset2),
            self.tree, self.tree, self.tree, self.tree
        ]


class QuickhullTask(Task):
    algorithm = quickhull
    grader_class = QuickhullGrader
    answers_class = QuickhullAnswers
    given_parser_class = PointListGivenJSONParser


from pytest import fixture
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.preparata import PreparataNode, PreparataThreadedBinTree 
from AlgoGrade.adapters import PointPydanticAdapter
from AlgoGrade.preparata import PreparataNodePydanticAdapter, PreparataThreadedBinTreePydanticAdapter


@fixture
def preparata_node_adapter():
    point_adapter = PointPydanticAdapter(coords=(1, 1))
    return PreparataNodePydanticAdapter(data=point_adapter)


@fixture
def preparata_node_regular():
    point = Point(1, 1)
    return PreparataNode(point)


def test_preparata_node_adapter(preparata_node_adapter, preparata_node_regular):
    assert preparata_node_adapter.regular_object() == preparata_node_regular
    assert PreparataNodePydanticAdapter.from_regular_object(preparata_node_regular) == preparata_node_adapter


def test_preparata_node_adapter_serialization(preparata_node_adapter):
    serialized_node = preparata_node_adapter.model_dump()
    deserialized_node = PreparataNodePydanticAdapter(**serialized_node)
    assert deserialized_node.regular_object() == preparata_node_adapter.regular_object()


def test_preparata_tree_adapter(preparata_node_adapter, preparata_node_regular):
    tree_adapter = PreparataThreadedBinTreePydanticAdapter(root=preparata_node_adapter)
    regular_object = PreparataThreadedBinTree(preparata_node_regular)

    assert tree_adapter.regular_object() == regular_object
    assert PreparataThreadedBinTreePydanticAdapter.from_regular_object(regular_object) == tree_adapter


def test_preparata_tree_adapter_serialization(preparata_node_adapter):
    tree_adapter = PreparataThreadedBinTreePydanticAdapter(root=preparata_node_adapter)
    serialized_tree = tree_adapter.model_dump()
    deserialized_tree = PreparataThreadedBinTreePydanticAdapter(**serialized_tree)
    assert deserialized_tree.regular_object().root.data == tree_adapter.regular_object().root.data


from math import isclose
from copy import deepcopy
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.quickhull import QuickhullNode, QuickhullTree
from AlgoGrade.quickhull import QuickhullGrader, QuickhullTask, QuickhullAnswers
from AlgoGrade.adapters import pycga_to_pydantic
from AlgoGrade.core import Scoring


points = [
    Point(0, 6),
    Point(8, 11),
    Point(10, 4),
    Point(7, 13),
    Point(6, 3),
    Point(3, 0),
    Point(4, 2),
    Point(12, 1),
    Point(14, 10),
    Point(5, 9),
    Point(3, 11),
    Point(1, 4),
]
givens = (points,)
hull = [points[0], points[10], points[3], points[8], points[7], points[5]]
task_class = QuickhullTask
scorings = [
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25, repeat_fine=0.5),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=1, fine=1)
]
correct_pycga_answers = task_class.solve_as_pycga_list(givens)
correct_pydantic_answers = task_class.solve_as_pydantic_list(givens)
correct_answers_wrapper = task_class.solve_as_answers_wrapper(givens)


def test_quickhull_grader_all_correct():
    tree = QuickhullTree(
        QuickhullNode(
            [
                points[0],
                points[10],
                points[9],
                points[3],
                points[1],
                points[8],
                points[7],
                points[2],
                points[4],
                points[6],
                points[5],
                points[11],
            ],
            subhull=hull
        )
    )

    tree.root.left = QuickhullNode(
        [points[0], points[10], points[9], points[3], points[1], points[8]],
        h=points[3],
        subhull=[points[0], points[10], points[3], points[8]]
    )
    tree.root.right = QuickhullNode(
        [points[8], points[7], points[2], points[4], points[6], points[5], points[11], points[0]],
        h=points[7],
        subhull=[points[8], points[7], points[5], points[0]]
    )

    tree.root.left.left = QuickhullNode([points[0], points[10], points[3]], h=points[10], subhull=[points[0], points[10], points[3]])
    tree.root.left.right = QuickhullNode([points[3], points[8]], subhull=[points[3], points[8]])
    tree.root.left.left.left = QuickhullNode([points[0], points[10]], subhull=[points[0], points[10]])
    tree.root.left.left.right = QuickhullNode([points[10], points[3]], subhull=[points[10], points[3]])

    tree.root.right.left = QuickhullNode([points[8], points[7]], subhull=[points[8], points[7]])
    tree.root.right.right = QuickhullNode(
        [points[7], points[4], points[6], points[5], points[11], points[0]],
        h=points[5],
        subhull=[points[7], points[5], points[0]]
    )
    tree.root.right.right.left = QuickhullNode([points[7], points[5]], subhull=[points[7], points[5]])
    tree.root.right.right.right = QuickhullNode([points[5], points[0]], subhull=[points[5], points[0]])

    leftmost_point, rightmost_point = points[0], points[8]
    s1, s2 = tree.root.left.points, tree.root.right.points

    pycga_answers = [(leftmost_point, rightmost_point, s1, s2), tree, tree, tree, tree]
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2)


def test_quickhull_grader_incorrect_first_step():
    pycga_answers = deepcopy(correct_pycga_answers)
    first_step_list = list(pycga_answers[0])
    first_step_list[0] = Point(100, 100)
    pycga_answers[0] = tuple(first_step_list)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_quickhull_grader_incorrect_h_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.h = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_quickhull_grader_incorrect_h_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1].root.h = Point(100, 100)
    pycga_answers[1].root.left.h = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.5)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.5)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.5)


def test_quickhull_grader_incorrect_points():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[2].root.left.left.points[0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_quickhull_grader_incorrect_finalization():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[3].root.left.right.left = QuickhullNode([]) # leaf node w/ 2 points is now not a leaf

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.25) # this test also triggers four node-related gradings: 0.25, 0.25, 0.25, 1

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.25)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.25)


def test_quickhull_grader_incorrect_merge():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[4].root.left.right.subhull = [Point(100, 100)]

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = QuickhullAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = QuickhullGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1)

    total_grade, answer_grades = QuickhullGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1)

    total_grade, answer_grades = QuickhullGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1)



from PyCompGeomAlgorithms.core import Point
from AlgoGrade.adapters import pycga_to_pydantic, pydantic_to_pycga, PointPydanticAdapter


def test_pycga_to_pydantic_type():
    assert pycga_to_pydantic(Point) == PointPydanticAdapter


def test_interlibrary_converters_pycga_object():
    point_pycga = Point(1, 2)
    point_adapter = PointPydanticAdapter(coords=(1, 2))

    assert pycga_to_pydantic(point_pycga) == point_adapter
    assert pydantic_to_pycga(point_adapter) == point_pycga


def test_interlibrary_converters_flat_dict():
    flat_dict_pycga = {Point(1, 2): 1, 42: "a", Point(3): 3, "abc": 1}
    flat_dict_adapter = {
        PointPydanticAdapter(coords=(1, 2)): 1,
        42: "a",
        PointPydanticAdapter(coords=(3,)): 3,
        "abc": 1
    }

    assert pycga_to_pydantic(flat_dict_pycga) == flat_dict_adapter
    assert pydantic_to_pycga(flat_dict_adapter) == flat_dict_pycga


def test_interlibrary_converters_nested_dict():
    nested_dict_pycga = {
        Point(1, 2): {1: 2, "abc": Point(3, 3), 42: {}},
        2: 2,
        "a": {Point(6, 5), 7}
    }
    nested_dict_adapter = {
        PointPydanticAdapter(coords=(1, 2)): {1: 2, "abc": PointPydanticAdapter(coords=(3, 3)), 42: {}},
        2: 2,
        "a": {PointPydanticAdapter(coords=(6, 5)), 7}
    }

    assert pycga_to_pydantic(nested_dict_pycga) == nested_dict_adapter
    assert pydantic_to_pycga(nested_dict_adapter) == nested_dict_pycga


def test_interlibrary_converters_flat_list():
    flat_list_pycga = [Point(1, 2), 42, Point(3), "abc"]
    flat_list_adapter = [PointPydanticAdapter(coords=(1, 2)), 42, PointPydanticAdapter(coords=(3,)), "abc"]

    assert pycga_to_pydantic(flat_list_pycga) == flat_list_adapter
    assert pydantic_to_pycga(flat_list_adapter) == flat_list_pycga


def test_interlibrary_converters_nested_list():
    nested_list_pycga = [
        {1: 2, "abc": Point(3, 3), 42: {}},
        2,
        [Point(6, 5), 7, []]
    ]
    nested_list_adapter = [
        {1: 2, "abc": PointPydanticAdapter(coords=(3, 3)), 42: {}},
        2,
        [PointPydanticAdapter(coords=(6, 5)), 7, []]
    ]

    assert pycga_to_pydantic(nested_list_pycga) == nested_list_adapter
    assert pydantic_to_pycga(nested_list_adapter) == nested_list_pycga



from typing import Any
from AlgoGrade.core import Scoring, Grader, Answers
from AlgoGrade.adapters import pycga_to_pydantic


class MockGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [
            cls.grade_object,
            cls.grade_object,
            cls.grade_iterable   
        ]


class MockAnswers(Answers):
    list: list[Any]

    @classmethod
    def from_iterable(cls, iterable):
        return cls(list=iterable)
    
    def to_pydantic_list(self):
        return self.list


scorings = [
    Scoring(max_grade=10, fine=5),
    Scoring(max_grade=10, fine=5),
    Scoring(min_grade=2, max_grade=10, fine=5, repeat_fine=7)
]


def test_grader1():
    pycga_answers = [1, 2, (3, 4)]
    correct_pycga_answers = pycga_answers

    total_grade, answer_grades = MockGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert total_grade == 30
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 10)]

    total_grade, answer_grades = MockGrader.grade_pydantic(pycga_to_pydantic(pycga_answers), pycga_to_pydantic(correct_pycga_answers), scorings)
    assert total_grade == 30
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 10)]

    total_grade, answer_grades = MockGrader.grade_answers_wrapper(
        MockAnswers.from_iterable(pycga_answers),
        MockAnswers.from_iterable(correct_pycga_answers),
        scorings
    )
    assert total_grade == 30
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 10)]


def test_grader2():
    pycga_answers = [1, 2, (3, 4)]
    correct_pycga_answers = [1, 3, (3, 4)]

    total_grade, answer_grades = MockGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert total_grade == 25
    assert answer_grades == [(1, 10), (2, 5), ((3, 4), 10)]

    total_grade, answer_grades = MockGrader.grade_pydantic(pycga_to_pydantic(pycga_answers), pycga_to_pydantic(correct_pycga_answers), scorings)
    assert total_grade == 25
    assert answer_grades == [(1, 10), (2, 5), ((3, 4), 10)]

    total_grade, answer_grades = MockGrader.grade_answers_wrapper(
        MockAnswers.from_iterable(pycga_answers),
        MockAnswers.from_iterable(correct_pycga_answers),
        scorings
    )
    assert total_grade == 25
    assert answer_grades == [(1, 10), (2, 5), ((3, 4), 10)]


def test_grader3():
    pycga_answers = [1, 2, (3, 4)]
    correct_pycga_answers = [1, 2, (0, 0)]

    total_grade, answer_grades = MockGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert total_grade == 23
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 3)]

    total_grade, answer_grades = MockGrader.grade_pydantic(pycga_to_pydantic(pycga_answers), pycga_to_pydantic(correct_pycga_answers), scorings)
    assert total_grade == 23
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 3)]

    total_grade, answer_grades = MockGrader.grade_answers_wrapper(
        MockAnswers.from_iterable(pycga_answers),
        MockAnswers.from_iterable(correct_pycga_answers),
        scorings
    )
    assert total_grade == 23
    assert answer_grades == [(1, 10), (2, 10), ((3, 4), 3)]
